#ifndef _ViewCell_H__
#define _ViewCell_H__

#include "Intersectable.h"

class Mesh;

//namespace GtpVisibilityPreprocessor {
  
class BspInterior;
class BspPvs;
/**
	View cell represented as a mesh
*/
class ViewCell: public Intersectable
{
public:
	/** Default constructor taking a mesh which represents the shape of
		the viewcell.
	*/
	ViewCell(Mesh *mesh): mMesh(mesh), mPvs(NULL) {}

	/** Returns pointer to the mesh which represents the shape of the viewcell.
	*/
	Mesh *GetMesh() {return mMesh;}

	/** Returns pointer to PVS. 
		@returns PVS, i.e., the visible BSP tree nodes.
	*/
	BspPvs *GetPVS() {return mPvs;}

	AxisAlignedBox3 GetBox()  {return mMesh->mBox;}
  
	int CastRay(Ray &ray) {return 0;}
  
	bool IsConvex() {return mMesh->mIsConvex;}
	bool IsWatertight() {return mMesh->mIsWatertight;}
	float IntersectionComplexity() {return mMesh->mFaces.size();}

	int Type() const { return VIEWCELL; }

	void GetRandomSurfacePoint(Vector3 &point, Vector3 &normal) {point = Vector3(0,0,0);};

protected:

	Mesh *mMesh;
	/// the PVS (i.e., the visible bsp tree nodes)
	BspPvs *mPvs;

};
  
//}; // GtpVisibilityPreprocessor

#endif
